@@ -40,6 +40,7 @@
  */
 
 EXTERN EXPORT(int,Pvm_Init) _ANSI_ARGS_((Tcl_Interp *interp));
+EXTERN EXPORT(int,Pvm_SafeInit) _ANSI_ARGS_((Tcl_Interp *interp));
 
 /*
  * Default directory in which to look for libraries:
@@ -55,34 +56,80 @@ static char defaultLibraryDir[200] = PVM_LIBRARY;
  * initialization.
  */
 
-static char *initScript =
-"proc init {} {\n\
+static char initScript[] =
+"proc tkpvmInit {} {\n\
     global pvm_library tkpvm_version tkpvm_patchLevel env\n\
-    rename init {}\n\
-    catch {set dirs $pvm_library}\n\
+    rename tkpvmInit {}\n\
+    set dirs {}\n\
     if [info exists env(PVM_LIBRARY)] {\n\
 	lappend dirs $env(PVM_LIBRARY)\n\
     }\n\
-    lappend dirs [file dirname [file dirname [info nameofexecutable]]]/lib/tkpvm$tkpvm_version\n\
-    if [string match {*[ab]*} $tkpvm_patchLevel] {\n\
-	set lib tkpvm$tkpvm_patchLevel\n\
-    } else {\n\
+    if [info exists env(EXT_FOLDER)] {\n\
+	lappend dirs [file join $env(EXT_FOLDER) \"Tool Command Language\" lib tkpvm$tkpvm_version]\n\
+    }\n\
+    lappend dirs $pvm_library\n\
+    set pvm_library {}\n\
+    if ![catch {uplevel #0 source -rsrc pvm}] {\n\
+	return\n\
+    }\n\
+    lappend dirs [file join [file dirname [info library]] tkpvm$tkpvm_version]\n\
+    set parentDir [file dirname [file dirname [info nameofexecutable]]]\n\
+    lappend dirs [file join $parentDir lib tkpvm$tkpvm_version]\n\
+    if {![regexp {.*[ab][12345]} $tkpvm_patchLevel lib]} {\n\
 	set lib tkpvm$tkpvm_version\n\
     }\n\
-    lappend dirs [file dirname [file dirname [pwd]]]/$lib/library\n\
-    lappend dirs [file dirname [pwd]]/library\n\
+    lappend dirs [file join [file dirname $parentDir] $lib library]\n\
+    lappend dirs [file join $parentDir library]\n\
     foreach i $dirs {\n\
-	set tkpvm_library $i\n\
-	if ![catch {uplevel #0 source $i/pvm.tcl}] {\n\
+	set pvm_library $i\n\
+	if ![catch {uplevel #0 source [list [file join $i pvm.tcl]]}] {\n\
 	    return\n\
 	}\n\
     }\n\
+    set pvm_library {}\n\
     set msg \"Can't find a usable pvm.tcl in the following directories: \n\"\n\
     append msg \"    $dirs\n\"\n\
     append msg \"This probably means that Tkpvm wasn't installed properly.\n\"\n\
     error $msg\n\
 }\n\
-init";
+tkpvmInit";
+
+static char safeInitScript[] =
+"proc tkpvmInit {} {\n\
+    global pvm_library tkpvm_version tkpvm_patchLevel env\n\
+    rename tkpvmInit {}\n\
+    set dirs {}\n\
+    if [info exists env(PVM_LIBRARY)] {\n\
+	lappend dirs $env(PVM_LIBRARY)\n\
+    }\n\
+    if [info exists env(EXT_FOLDER)] {\n\
+	lappend dirs [file join $env(EXT_FOLDER) \"Tool Command Language\" lib tkpvm$tkpvm_version]\n\
+    }\n\
+    lappend dirs $pvm_library\n\
+    set pvm_library {}\n\
+    if ![catch {uplevel #0 source -rsrc pvm}] {\n\
+	return\n\
+    }\n\
+    lappend dirs [file join [file dirname [info library]] tkpvm$tkpvm_version]\n\
+    set parentDir [file dirname [file dirname [info nameofexecutable]]]\n\
+    lappend dirs [file join $parentDir lib tkpvm$tkpvm_version]\n\
+    if {![regexp {.*[ab][12345]} $tkpvm_patchLevel lib]} {\n\
+	set lib tkpvm$tkpvm_version\n\
+    }\n\
+    lappend dirs [file join [file dirname [file dirname [info library]]] $lib library]\n\
+    foreach i $dirs {\n\
+	set pvm_library $i\n\
+	if ![catch {uplevel #0 source [list [file join $i pvm.tcl]]}] {\n\
+	    return\n\
+	}\n\
+    }\n\
+    set pvm_library {}\n\
+    set msg \"Can't find a usable pvm.tcl in the following directories: \n\"\n\
+    append msg \"    $dirs\n\"\n\
+    append msg \"This probably means that Tkpvm wasn't installed properly.\n\"\n\
+    error $msg\n\
+}\n\
+tkpvmInit";
 
 static char **initScripts = NULL;
 
@@ -91,21 +138,31 @@ typedef struct {
     int (*cmdProc) _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
 	    int argc, char **argv));
 				/* Command procedure. */
+    int flags;			/* If flag 0 is set, this command will be
+                                 * exposed in a safe interpreter. Otherwise
+                                 * it will be hidden in a safe interpreter.
+				 * If flag 1 is set, this command is availabe
+				 * as is, otherwise it is only availabe as
+				 * option for the "pvm" command. */
 } TkpvmCmd;
 
 static TkpvmCmd commands[] = {
-    {"addhost",		Pvm_AddhostCmd},
-    {"delhost",		Pvm_DelhostCmd},
-    {"halt",		Pvm_HaltCmd},
-    {"tid",		Pvm_TidCmd},
-    {"joingroup",	Pvm_JoingroupCmd},
-    {"pvm",		Pvm_PvmCmd},
-    {"leavegroup",	Pvm_LeavegroupCmd},
-    {"parent",		Pvm_ParentCmd},
-    {"recv",		Pvm_RecvCmd},
-    {"spawn",		Pvm_SpawnCmd},
-    {"tasks",		Pvm_TasksCmd},
-    {(char *) NULL,	(int (*)()) NULL}
+    {"addhost",		Pvm_AddhostCmd,		2},
+    {"bind",		Pvm_BindCmd,		1},
+    {"conf",		Pvm_ConfCmd,		3},
+    {"delhost",		Pvm_DelhostCmd,		2},
+    {"halt",		Pvm_HaltCmd,		2},
+    {"joingroup",	Pvm_JoingroupCmd,	3},
+    {"kill",		Pvm_KillCmd,		0},
+    {"leavegroup",	Pvm_LeavegroupCmd,	3},
+    {"parent",		Pvm_ParentCmd,		3},
+    {"pvm",		Pvm_PvmCmd,		3},
+    {"recv",		Pvm_RecvCmd,		3},
+    {"send",		Pvm_SendCmd,		1},
+    {"spawn",		Pvm_SpawnCmd,		2},
+    {"tasks",		Pvm_TasksCmd,		3},
+    {"tid",		Pvm_TidCmd,		3},
+    {(char *) NULL,	(int (*) _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **))) NULL}
 };
 
 /*
@@ -134,8 +191,12 @@ Pvm_PvmCmd(dummy, interp, argc, argv)
     char **argv;			/* Argument strings. */
 {
     size_t length;
-    register char *p;
-    int c;
+    char *temp, *argv1;
+    TkpvmCmd *cmdPtr;
+
+    int c, isSafe, result;
+    int (*optfn) _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **)) =
+	    (int (*) _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **))) NULL;
 
     if (argc < 2) {
 	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
@@ -144,41 +205,34 @@ Pvm_PvmCmd(dummy, interp, argc, argv)
     }
     c = argv[1][0];
     length = strlen(argv[1]);
-    if ((c == 'a') && (strncmp(argv[1], "addhost", length) == 0)) {
-	return Pvm_AddhostCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'b') && (strncmp(argv[1], "bind", length) == 0)) {
-	return Pvm_BindCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'd') && (strncmp(argv[1], "delhost", length) == 0)) {
-	return Pvm_DelhostCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'h') && (strncmp(argv[1], "halt", length) == 0)) {
-	return Pvm_HaltCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 't') && (strncmp(argv[1], "tid", length) == 0)) {
-	return Pvm_TidCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'j') && (strncmp(argv[1], "joingroup", length) == 0)) {
-	return Pvm_JoingroupCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'k') && (strncmp(argv[1], "kill", length) == 0)) {
-	return Pvm_KillCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'l') && (strncmp(argv[1], "leavegroup", length) == 0)) {
-	return Pvm_LeavegroupCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'p') && (strncmp(argv[1], "parent", length) == 0)) {
-	return Pvm_ParentCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 'r') && (strncmp(argv[1], "recv", length) == 0)) {
-	return Pvm_RecvCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 's') && (strncmp(argv[1], "send", length) == 0)
-	    && (length >= 2)) {
-	return Pvm_SendCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 's') && (strncmp(argv[1], "spawn", length) == 0)
-	    && (length >= 2)) {
-	return Pvm_SpawnCmd(dummy, interp, argc-1, argv+1);
-    } else if ((c == 't') && (strncmp(argv[1], "tasks", length) == 0)) {
-	return Pvm_TasksCmd(dummy, interp, argc-1, argv+1);
-    } else {
-	Tcl_AppendResult(interp, "bad option \"", argv[1],
-		"\": should be addhost, bind, delhost, halt, tid, ",
-		"joingroup, kill, leavegroup, parent, recv, send, ",
-		"spawn or tasks", (char *) NULL);
-	return TCL_ERROR;
+    isSafe = Tcl_IsSafe(interp);
+    for (cmdPtr = commands; cmdPtr->name != NULL; cmdPtr++) {
+	if (((!isSafe) || (cmdPtr->flags & 1)) &&
+		!strncmp(cmdPtr->name, argv[1], length)) {
+	    if (optfn != NULL) {
+		optfn = NULL;
+		break;
+	    }
+	    optfn = cmdPtr->cmdProc;
+	}
     }
+    if (optfn != NULL) {
+	temp = (char *) ckalloc (strlen(argv[0]) + strlen(argv[1]) + 2);
+	strcpy(temp,argv[0]);
+	strcat(temp," ");
+	strcat(temp,argv[1]);
+	argv1 = argv[1];
+	argv[1] = temp;
+	result = (* optfn) (dummy, interp, argc-1, argv+1);
+	argv[1] = argv1;
+	ckfree(temp);
+	return result;
+    }
+    Tcl_AppendResult(interp, "bad or ambigious option \"", argv[1],
+	    "\": should be addhost, bind, delhost, halt, ",
+	    "joingroup, kill, leavegroup, parent, recv, send, ",
+	    "spawn, tasks or tid,", (char *) NULL);
+    return TCL_ERROR;
 }
 
 /*
@@ -306,6 +360,55 @@ Pvm_HaltCmd(dummy, interp, argc, argv)
     return TCL_OK;
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * Pvm_ConfCmd --
+ *
+ *	This procedure is invoked to process the "conf" Tkpvm command.
+ *	See the user documentation for details on what it does.
+ *
+ * Results:
+ *	A standard Tcl result.
+ *
+ * Side effects:
+ *	See the user documentation.
+ *
+ *----------------------------------------------------------------------
+ */
+
+	/* ARGSUSED */
+int
+Pvm_ConfCmd(dummy, interp, argc, argv)
+    ClientData dummy;			/* Not used. */
+    Tcl_Interp *interp;			/* Current interpreter. */
+    int argc;				/* Number of arguments. */
+    char **argv;			/* Argument strings. */
+{
+    int no_of_hosts;
+    int arch_info;
+    struct pvmhostinfo *host_info;
+    int i;
+    char buffer[20];
+
+    if (argc != 1) {
+	Tcl_AppendResult(interp, "wrong # args: should be \"",
+		argv[0], "\"", (char *) NULL);
+	return TCL_ERROR;
+    }
+    if (!pvm_config(&no_of_hosts, &arch_info, &host_info)) {
+	for (i = 0; i < no_of_hosts; i++) {
+	    sprintf(buffer," {0x%x",host_info[i].hi_tid);
+	    Tcl_AppendResult(interp, buffer + (i==0), (char *) NULL);
+	    Tcl_AppendElement(interp, host_info[i].hi_name);
+	    Tcl_AppendElement(interp, host_info[i].hi_arch);
+	    sprintf(buffer," %d}",host_info[i].hi_speed);
+	    Tcl_AppendResult(interp, buffer, (char *) NULL);
+	}
+    }
+    return (TCL_OK);
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -538,7 +641,7 @@ DllEntryPoint(hInst, reason, reserved)
 /*
  *--------------------------------------------------------------
  *
- * Pvm_Init --
+ * Pvm_Init , Pvm_SafeInit--
  *	Create pvm commands.
  *
  * Results:
@@ -556,6 +659,12 @@ EXPORT(int,Pvm_Init)(interp)
     register char *libDir;
     register char *p;
     register TkpvmCmd *cmdPtr;
+#ifdef ITCL_NAMESPACES
+    Itcl_Namespace spaceId;	/* Token for "pvm" namespace created, used
+				 * to delete the namespace on errors. */
+    char cmdPath[200];
+#endif
+    int isSafe;
 
     if (Tcl_PkgRequire(interp, "Tcl", TCL_VERSION, 0) == NULL) {
 	return TCL_ERROR;
@@ -563,7 +672,6 @@ EXPORT(int,Pvm_Init)(interp)
     if (Tcl_PkgProvide(interp, "Pvm", TKPVM_VERSION) != TCL_OK) {
 	return TCL_ERROR;
     }
-
     /*
      * Set variables for the intepreter.
      */
@@ -578,14 +686,32 @@ EXPORT(int,Pvm_Init)(interp)
 	}
     }
 
-    for (cmdPtr = commands; cmdPtr->name != NULL; cmdPtr++) {
-	Tcl_CreateCommand(interp, cmdPtr->name, cmdPtr->cmdProc,
-		(ClientData) NULL, (void (*)()) NULL);
+    isSafe = Tcl_IsSafe(interp);
+
+#ifdef ITCL_NAMESPACES
+    if (Itcl_CreateNamesp(interp, "pvm", (ClientData)0, (Itcl_DeleteProc *)0,
+		  &spaceId) != TCL_OK) {
+	return TCL_ERROR;
     }
+    for (cmdPtr = commands; cmdPtr->name != NULL; cmdPtr++) {
+	if ((cmdPtr->flags & 2) && ((!isSafe) || (cmdPtr->flags & 1))) {
+	    sprintf(cmdPath, "pvm::%s", cmdPtr->name);
+	    Tcl_CreateCommand(interp, cmdPath, cmdPtr->cmdProc,
+		    (ClientData) NULL, (void (*) _ANSI_ARGS_((ClientData))) NULL);
+	}
+    }
+#else
+    for (cmdPtr = commands; cmdPtr->name != NULL; cmdPtr++) {
+	if ((cmdPtr->flags & 2) && ((!isSafe) || (cmdPtr->flags & 1))) {
+	    Tcl_CreateCommand(interp, cmdPtr->name, cmdPtr->cmdProc,
+		    (ClientData) NULL, (void (*) _ANSI_ARGS_((ClientData))) NULL);
+	}
+    }
+#endif
 
     libDir = Tcl_GetVar(interp,"pvm_library",TCL_GLOBAL_ONLY);
     if (libDir == NULL) {
-	Tcl_SetVar(interp, "pvm_library",   defaultLibraryDir,		TCL_GLOBAL_ONLY);
+	Tcl_SetVar(interp, "pvm_library", defaultLibraryDir,	TCL_GLOBAL_ONLY);
     }
 
     Tcl_SetVar(interp, "pvm_version",     pvm_version(),	TCL_GLOBAL_ONLY);
@@ -600,17 +726,32 @@ EXPORT(int,Pvm_Init)(interp)
 	Tcl_DStringInit(&data);
 	while(*p) {
 	    /* Copy the constant into a dynamic string. This */
-	    /* is necessary because Tcl7.5 doesn't accept    */
+	    /* is necessary because Tcl doesn't accept    */
 	    /* constants as an argument to Tcl_Eval()        */
 	    Tcl_DStringSetLength(&data,0);
 	    Tcl_DStringAppend(&data,*p++,-1);
 	    if(Tcl_Eval(interp,Tcl_DStringValue(&data)) == TCL_ERROR) {
 		Tcl_DStringFree(&data);
+#ifdef ITCL_NAMESPACES
+		Itcl_DeleteNamesp(spaceId);
+#endif
 		return TCL_ERROR;
 	    }
 	}
 	Tcl_DStringFree(&data);
 	return TCL_OK;
     }
-    return Tcl_Eval(interp, initScript);
+    if (Tcl_Eval(interp, initScript) != TCL_OK) {
+#ifdef ITCL_NAMESPACES
+	Itcl_DeleteNamesp(spaceId);
+#endif
+	return TCL_ERROR;
+    }
+    return TCL_OK;
+}
+
+EXPORT(int,Pvm_SafeInit)(interp)
+    Tcl_Interp *interp;
+{
+    return Pvm_Init(interp);
 }
