@@ -1,4 +1,6 @@
-#!/usr/local/bin/pvmsh1.0
+#!/usr/local/bin/pvmsh1.1
+
+package require Pvm 1.1
 
 set result ""
 set parent [parent]
@@ -10,13 +12,7 @@ proc bgerror {msg} {
 
 bind $parent 1 {
     set type [recv string]
-    set result $type
-    set counter [recv int]
-    lappend result $counter
-    while {$counter>0} {
-	lappend result [recv $type]
-	incr counter -1;
-    }
+    set result "$type [eval recv $type]"
     pvm send $parent 0 -string $result
 }
 
